

import numpy as np
import datastock as ds


import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from mpl_toolkits.mplot3d import Axes3D


from .. import version


_GITHUB = "https://github.com/ToFuProject/tofu"
_WINDEF = f"tofu {version.__version__} - report issues at {_GITHUB}"


_DFS = {
    1: (8, 5),
    2: (11, 5),
    3: (13, 9),
    4: (15, 9),
}


# #################################################################
# #################################################################
#               proj
# #################################################################


def _proj(
    proj=None,
    pall=None,
):

    if pall is None:
        pall = ['cross', 'hor', '3d', 'camera']

    # proj
    if proj == 'all':
        proj = pall

    if isinstance(proj, str):
        proj = [proj]

    proj = ds._generic_check._check_var_iter(
        proj, 'proj',
        default=pall,
        allowed=pall,
        types=list,
        types_iter=str,
    )
    return proj


# #################################################################
# #################################################################
#               Default axes
# #################################################################


# Generic
def get_dax_diag(
    proj=None,
    dmargin=None,
    fs=None,
    tit=None,
    wintit=None,
    is2d=None,
):

    # ----------------------
    # Check inputs

    # fs
    fs = ds._generic_check._check_var(
        fs, 'fs',
        types=tuple,
        default=_DFS[len(proj)],
    )

    # tit
    tit = ds._generic_check._check_var(
        tit, 'tit',
        types=str,
        default='',
    )

    # wintit
    wintit = ds._generic_check._check_var(
        wintit, 'wintit',
        types=str,
        default=_WINDEF,
    )

    # is2d
    is2d = ds._generic_check._check_var(
        is2d, 'is2d',
        types=bool,
        default=False,
    )

    # -------------
    # Create figure

    fig = plt.figure(figsize=fs)
    fig.canvas.set_window_title(wintit)
    fig.suptitle(tit, size=12, fontweight='bold')

    # ------------
    # Populate dax

    dax = {}

    # Populate with default axes if necessary
    if len(proj) == 1:
        lax = _ax_single(
            fig=fig,
            dmargin=dmargin,
            proj=proj,
        )

    elif len(proj) == 2:
        lax = _ax_double(
            fig=fig,
            dmargin=dmargin,
            proj=proj,
        )

    elif len(proj) == 3:
        lax = _ax_3(
            fig=fig,
            dmargin=dmargin,
            proj=proj,
        )
    else:
        lax = _ax_4(
            fig=fig,
            dmargin=dmargin,
            proj=proj,
        )

    for ii, ax in enumerate(lax):
        _ax_set(ax=ax, proj=proj[ii], is2d=is2d)
        dax[proj[ii]] = ax

    return dax


# #################################################################
# #################################################################
#               Default axes
# #################################################################


def _ax_single(
    fig=None,
    dmargin=None,
    proj=None,
):

    # ------------
    # check inputs

    # dmargin
    dmargin = ds._generic_check._check_var(
        dmargin, 'dmargin',
        types=dict,
        default={
            'bottom': 0.12, 'top': 0.9,
            'left': 0.10, 'right': 0.95,
        },
    )

    # ----------------------
    # create figure and axes

    gs = gridspec.GridSpec(ncols=1, nrows=1, **dmargin)

    if proj[0] == '3d':
        ax = fig.add_subplot(gs[0, 0], projection='3d')
    else:
        ax = fig.add_subplot(gs[0, 0])

    return [ax]


def _ax_double(
    fig=None,
    dmargin=None,
    proj=None,
):

    # ------------
    # check inputs

    # dmargin
    dmargin = ds._generic_check._check_var(
        dmargin, 'dmargin',
        types=dict,
        default={
            'bottom': 0.12, 'top': 0.9,
            'left': 0.08, 'right': 0.98,
            'wspace': 0.20, 'hspace': 0.05,
        },
    )

    # ----------------------
    # create figure and axes

    gs = gridspec.GridSpec(ncols=2, nrows=1, **dmargin)

    lax = []
    for ii, pp in enumerate(proj):
        if pp == '3d':
            lax.append(fig.add_subplot(gs[0, ii], projection='3d'))
        else:
            lax.append(fig.add_subplot(gs[0, ii]))

    return lax


def _ax_3(
    fig=None,
    dmargin=None,
    proj=None,
):

    # ------------
    # check inputs

    # dmargin
    dmargin = ds._generic_check._check_var(
        dmargin, 'dmargin',
        types=dict,
        default={
            'bottom': 0.12, 'top': 0.9,
            'left': 0.06, 'right': 0.98,
            'wspace': 0.10, 'hspace': 0.05,
        },
    )

    # ----------------------
    # create figure and axes

    gs = gridspec.GridSpec(ncols=2, nrows=2, **dmargin)

    lax = []
    for ii, pp in enumerate(proj):
        if pp == '3d':
            lax.append(fig.add_subplot(gs[0, ii], projection='3d'))
        else:
            lax.append(fig.add_subplot(gs[ii % 2, ii // 2]))

    return lax


def _ax_4(
    fig=None,
    dmargin=None,
    proj=None,
):

    # ------------
    # check inputs

    # dmargin
    dmargin = ds._generic_check._check_var(
        dmargin, 'dmargin',
        types=dict,
        default={
            'bottom': 0.08, 'top': 0.90,
            'left': 0.06, 'right': 0.96,
            'wspace': 0.20, 'hspace': 0.20,
            'width_ratios': [0.6, 0.4],
            'height_ratios': [0.4, 0.6],
        },
    )

    # ----------------------
    # create figure and axes

    gs = gridspec.GridSpec(ncols=2, nrows=2, **dmargin)
    dgs = {
        'cross': (0, 0),
        'hor': (0, 1),
        '3d': (1, 0),
        'camera': (1, 1),
    }

    lax = []
    for ii, pp in enumerate(proj):
        if pp == '3d':
            lax.append(fig.add_subplot(
                gs[dgs[pp][0], dgs[pp][1]],
                projection='3d',
            ))
        else:
            lax.append(fig.add_subplot(
                gs[dgs[pp][0], dgs[pp][1]],
            ))

    return lax


def _ax_set(ax=None, proj=None, is2d=None):

    if proj == 'cross':

        ax.set_xlabel('R (m)')
        ax.set_ylabel('Z (m)')
        ax.set_aspect('equal', adjustable='datalim')

    elif proj == 'hor':

        ax.set_xlabel('X (m)')
        ax.set_ylabel('Y (m)')
        ax.set_aspect('equal', adjustable='datalim')

    elif proj == '3d':

        ax.set_xlabel('X (m)')
        ax.set_ylabel('Y (m)')
        ax.set_zlabel('Z (m)')
        # ax.set_aspect('equal')

    elif proj == 'camera':

        if is2d:
            ax.set_xlabel('x0 (m)')
            ax.set_ylabel('x1 (m)')
            ax.set_aspect('equal', adjustable='datalim')
        else:
            ax.set_xlabel('ind')

    return



# -*- coding: utf-8 -*-
"""
Provide data handling class and methods (storing, processing, plotting...)
"""
from ._core import *
from ._DataCollection_class1_interactivity import *
from ._spectrallines_class import *
from ._class1_Plasma2D import *
from ._class8_Diagnostic import *
from ._inversions_checks import get_available_inversions_algo



# -*- coding: utf-8 -*-


# Built-in
import copy


# Common
import numpy as np
import datastock as ds


# #############################################################################
# #############################################################################
#                           Matrix - compute
# #############################################################################


def _compute_check(
    coll=None,
    key=None,
    key_chan=None,
    nlos=None,
    method=None,
    resMode=None,
    crop=None,
    name=None,
    store=None,
    verb=None,
):

    # key
    lk = list(coll.dobj.get('bsplines', {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lk,
    )

    # key_chan
    lk = (
        [k0 for k0, v0 in coll.dref.items() if v0['size'] == nlos]
        + [None]
    )
    key_chan = ds._generic_check._check_var(
        key_chan, 'key_chan',
        allowed=lk,
    )

    # method
    method = ds._generic_check._check_var(
        method, 'method',
        default='los',
        types=str,
        allowed=['los'],
    )

    # resMode
    resMode = ds._generic_check._check_var(
        resMode, 'resMode',
        default='abs',
        types=str,
        allowed=['abs', 'rel'],
    )

    # crop
    crop = ds._generic_check._check_var(
        crop, 'crop',
        default=True,
        types=bool,
    )
    crop = crop and coll.dobj['bsplines'][key]['crop'] not in [None, False]

    # name
    if name is None:
        lmat = [
            kk for kk in coll.dobj.get('matrix', {}).keys()
            if kk.startswith('matrix')
        ]
        name = f'matrix{len(lmat)}'
    c0 = (
        isinstance(name, str)
        and name not in coll.dobj.get('matrix', {}).keys()
    )
    if not c0:
        msg = (
            "Arg name must be a str not already taken!\n"
            f"\t- already taken: {coll.dobj.get('matrix', {}).keys()}\n"
            f"\t- provided: {name}"
        )
        raise Exception(msg)

    # store
    store = ds._generic_check._check_var(
        store, 'store',
        default=True,
        types=bool,
    )

    # verb
    if verb is None:
        verb = True
    if not isinstance(verb, bool):
        msg = (
            f"Arg verb must be a bool!\n"
            f"\t- provided: {verb}"
        )
        raise Exception(msg)

    return key, key_chan, method, resMode, crop, name, store, verb


def compute(
    coll=None,
    key=None,
    key_chan=None,
    cam=None,
    res=None,
    resMode=None,
    method=None,
    crop=None,
    name=None,
    store=None,
    verb=None,
):
    """ Compute the geometry matrix using:
            - a Plasma2DRect instance with a key to a bspline set
            - a cam instance with a resolution
    """

    # -----------
    # check input

    nlos = cam.nRays
    key, key_chan, method, resMode, crop, name, store, verb = _compute_check(
        coll=coll,
        key=key,
        key_chan=key_chan,
        nlos=nlos,
        method=method,
        resMode=resMode,
        crop=crop,
        name=name,
        store=store,
        verb=verb,
    )

    # -----------
    # prepare

    shapebs = coll.dobj['bsplines'][key]['shape']
    km = coll.dobj['bsplines'][key]['mesh']
    mtype = coll.dobj[coll._which_mesh][km]['type']

    # prepare indices
    indbs = coll.select_ind(
        key=key,
        returnas=bool,
        crop=crop,
    )

    # prepare matrix
    is3d = False
    if mtype == 'polar':
        radius2d = coll.dobj[coll._which_mesh][km]['radius2d']
        r2d_reft = coll.get_time(key=radius2d)[2]
        if r2d_reft is not None:
            r2d_nt = coll.dref[r2d_reft]['size']
            if r2d_nt > 1:
                shapemat = tuple(np.r_[r2d_nt, nlos, indbs.sum()])
                is3d = True

    if not is3d:
        shapemat = tuple(np.r_[nlos, indbs.sum()])

    mat = np.zeros(shapemat, dtype=float)

    # -----------
    # compute

    if method == 'los':
        # discretize lines once, then evaluated at points
        pts, reseff, ind = cam.get_sample(
            res=res,
            resMode=resMode,
            DL=None,
            method='sum',
            ind=None,
            pts=True,
            compact=True,
            num_threads=10,
            Test=True,
        )
        lr = np.split(np.hypot(pts[0, :], pts[1, :]), ind)
        lz = np.split(pts[2, :], ind)

        if verb:
            nmax = len(f"Geometry matrix for {key}, channel {nlos} / {nlos}")
            nn = 10**(np.log10(nlos)-1)

        for ii in range(nlos):

            # verb
            if verb:
                msg = f"Geom. matrix for {key}, chan {ii+1} / {nlos}"
                end = '\n' if ii == nlos-1 else '\r'
                print(msg.ljust(nmax), end=end, flush=True)

            # compute
            mati = coll.interpolate_profile2d(
                key=key,
                R=lr[ii],
                Z=lz[ii],
                grid=False,
                indbs=indbs,
                details=True,
                nan0=False,
                nan_out=False,
                return_params=False,
            )[0]
            assert mati.ndim in [2, 3], mati.shape

            # integrate
            if is3d:
                mat[:, ii, :] = np.nansum(mati, axis=1) * reseff[ii]
            elif mati.ndim == 3 and mati.shape[0] == 1:
                mat[ii, :] = np.nansum(mati[0, ...], axis=0) * reseff[ii]
            else:
                mat[ii, :] = np.nansum(mati, axis=0) * reseff[ii]

        # scpintg.simps(val, x=None, axis=-1, dx=loc_eff_res[0])

    # -----------
    # return

    if store:

        # add key chan if necessary
        dref = None
        if key_chan is None:
            lrchan = [
                k0 for k0, v0 in coll.dref.items()
                if k0.startswith('chan') and k0[4:].isdecimal()
            ]
            if len(lrchan) == 0:
                chann = 0
            else:
                chann = max([int(k0.replace('chan', '')) for k0 in lrchan]) + 1
            key_chan = f'chan{chann}'

            dref = {
                key_chan: {
                    'data': np.arange(0, nlos),
                },
            }

        # add matrix data
        keycropped = coll.dobj['bsplines'][key]['ref-bs'][0]
        if crop is True:
            keycropped = f'{keycropped}-crop'

        # ref
        if is3d:
            ref = (r2d_reft, key_chan, keycropped)
        else:
            ref = (key_chan, keycropped)

        # add data
        ddata = {
            name: {
                'data': mat,
                'ref': ref,
            },
        }

        # add matrix obj
        dobj = {
            'matrix': {
                name: {
                    'bsplines': key,
                    'cam': cam.Id.Name,
                    'data': name,
                    'crop': crop,
                    'shape': mat.shape,
                },
            },
        }

        coll.update(dref=dref, ddata=ddata, dobj=dobj)

    else:
        return mat


# #############################################################################
# #############################################################################
#               retrofit                   
# #############################################################################


def _compute_retrofit_data_check(
    # resources
    coll=None,
    # inputs
    key=None,
    key_matrix=None,
    key_profile2d=None,
    t=None,
    # parameters
    store=None,
):

    #----------
    # keys

    # key
    lout = coll.ddata.keys()
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        excluded=lout,
    )

    # key_matrix
    lok = coll.dobj.get('matrix', {}).keys()
    key_matrix = ds._generic_check._check_var(
        key_matrix, 'key_mtrix',
        types=str,
        allowed=lok,
    )
    keybs = coll.dobj['matrix'][key_matrix]['bsplines']
    keym = coll.dobj['bsplines'][keybs]['mesh']
    mtype = coll.dobj[coll._which_mesh][keym]['type']

    nchan, nbs = coll.ddata[key_matrix]['data'].shape[-2:]
    refchan, refbs = coll.ddata[key_matrix]['ref'][-2:]

    # key_pofile2d
    lok = [
        k0 for k0, v0 in coll.ddata.items()
        if v0['bsplines'] == keybs
    ]
    key_profile2d = ds._generic_check._check_var(
        key_profile2d, 'key_profile2d',
        types=str,
        allowed=lok,
    )

    # time management
    hastime, reft, keyt, t_out, dind = coll.get_time_common(
        keys=[key_matrix, key_profile2d],
        t=t,
        ind_strict=False,
    )
    if hastime and t_out is not None and reft is None:
        reft = f'{key}-nt'
        keyt = f'{key}-t'

    ist_mat = coll.get_time(key=key_matrix)[0]
    ist_prof = coll.get_time(key=key_profile2d)[0]

    # reft, keyt and refs
    if hastime and t_out is not None:
        nt = t_out.size
        refs = (reft, refchan)
    else:
        nt = 0
        reft = None
        keyt = None
        refs = (refchan,)

    return (
        key, keybs, keym, mtype,
        key_matrix, key_profile2d,
        hastime, t_out, keyt, reft, refs,
        nt, nchan, nbs,
        ist_mat, ist_prof, dind,
    )


def compute_retrofit_data(
    # resources
    coll=None,
    # inputs
    key=None,
    key_matrix=None,
    key_profile2d=None,
    t=None,
    # parameters
    store=None,
):

    # ------------
    # check inputs

    (
        key, keybs, keym, mtype,
        key_matrix, key_profile2d,
        hastime, t, keyt, reft, refs,
        nt, nchan, nbs,
        ist_mat, ist_prof, dind,
    ) = _compute_retrofit_data_check(
        # resources
        coll=coll,
        # inputs
        key=key,
        key_matrix=key_matrix,
        key_profile2d=key_profile2d,
        t=t,
        # parameters
        store=store,
    )

    # --------
    # compute

    matrix = coll.ddata[key_matrix]['data']
    coefs = coll.ddata[key_profile2d]['data']
    if mtype == 'rect':
        indbs_tf = coll.select_bsplines(
            key=keybs,
            returnas='ind',
        )
        if hastime and ist_prof:
            coefs = coefs[:, indbs_tf[0], indbs_tf[1]]
        else:
            coefs = coefs[indbs_tf[0], indbs_tf[1]]

    if hastime:

        retro = np.full((nt, nchan, nbs), np.nan)

        # get time indices
        if ist_mat:
            if dind.get(key_matrix, {}).get('ind') is not None:
                imat = dind[key_matrix]['ind']
            else:
                imat = np.arange(nt)

        if ist_prof:
            if dind.get(key_profile2d, {}).get('ind') is not None:
                iprof = dind[key_profile2d]['ind']
            else:
                iprof = np.arange(nt)

        # compute matrix product
        if ist_mat and ist_prof:
            retro = np.array([
                matrix[imat[ii], :, :].dot(coefs[iprof[ii], :])
                for ii in range(nt)
            ])
        elif ist_mat:
            retro = np.array([
                matrix[imar[ii], :, :].dot(coefs)
                for ii in range(nt)
            ])
        elif ist_prof:
            retro = np.array([
                matrix.dot(coefs[iprof[ii], :])
                for ii in range(nt)
            ])
    else:
        retro = matrix.dot(coefs)

    # --------
    # store

    if store:

        # add data
        ddata = {
            key: {
                'data': retro,
                'ref': refs,
                'dim': None,
                'quant': None,
                'name': None,
            },
        }

        # add reft + t if new
        if hastime and keyt not in coll.ddata.keys():
            ddata[keyt] = {'data': t, 'ref': reft, 'dim': 'time'}
        if hastime and reft not in coll.dref.keys():
            dref = {reft: {'size': t.size}}
        else:
            dref = None

        # update
        coll.update(dref=dref, ddata=ddata)

    else:
        return retro, t, keyt, reft



# -*- coding: utf-8 -*-


# Built-in


# Common
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import datastock as ds


# specific
from . import _mesh_comp
from . import _generic_check
from . import _generic_plot


# ##################################################################
# ##################################################################
#                           plot check
# ##################################################################


def _plot_diagnostic_check(
    coll=None,
    key=None,
    # figure
    proj=None,
    data=None,
    rocking_curve=None,
    # interactivity
    color_dict=None,
    nlos=None,
    connect=None,
):

    # -------
    # key

    lok = list(coll.dobj.get('diagnostic', {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )
    optics, optics_cls = coll.get_diagnostic_optics(key)
    if 'crystal' in optics_cls:
        kcryst = optics[optics_cls.index('crystal')]
    else:
        kcryst = None

    # -----
    # proj

    proj = _generic_plot._proj(
        proj=proj,
        pall=['cross', 'hor', '3d', 'camera'],
    )

    # -------
    # data

    defdata = 'etendue'
    c0 = data is None and coll.dobj['diagnostic'][key].get(defdata) is None

    ref = None
    if not c0:
        lok = [
            k0 for k0, v0 in coll.ddata.items()
            if v0.get('camera') == coll.dobj['diagnostic'][key]['optics'][0]
        ]
        ldiag = [
            k0 for k0, v0 in coll.dobj['diagnostic'][key].items()
            if isinstance(v0, str) and v0 in lok
        ]
        klos = coll.dobj['diagnostic'][key].get('los')
        if klos is None:
            lrays = []
        else:
            lrays = [
                k0 for k0, v0 in coll.dobj['rays'][klos].items()
                if isinstance(v0, str) and v0 in lok
            ]

        if kcryst is None:
            llamb = []
        else:
            llamb = ['lamb', 'lambmin', 'lambmax', 'res']

        data = ds._generic_check._check_var(
            data, 'data',
            types=str,
            allowed=lok + ldiag + lrays + llamb,
            default=defdata,
        )

        if data in ldiag:
            data = coll.dobj['diagnostic'][key][data]
        elif data in lrays:
            data = coll.dobj['rays'][klos][data]
        elif data in llamb:
            data, ref = coll.get_diagnostic_lamb(
                key=key,
                rocking_curve=rocking_curve,
                lamb=data,
            )

    # -------
    # color_dict

    if color_dict is None:
        lc = ['r', 'g', 'b', 'm', 'c', 'y']
        color_dict = {
            'x': lc,
            'y': lc,
        }

    # -------
    # nlos

    nlos = ds._generic_check._check_var(
        nlos, 'nlos',
        types=int,
        default=5,
    )

    # -------
    # connect

    connect = ds._generic_check._check_var(
        connect, 'connect',
        types=bool,
        default=True,
    )

    return (
        key,
        proj,
        data,
        ref,
        color_dict,
        nlos,
        connect,
    )


# ##################################################################
# ##################################################################
#                           plot main
# ##################################################################


def _plot_diagnostic(
    coll=None,
    key=None,
    optics=None,
    elements=None,
    proj=None,
    los_res=None,
    # data plot
    data=None,
    cmap=None,
    vmin=None,
    vmax=None,
    # config
    plot_config=None,
    # figure
    dax=None,
    dmargin=None,
    fs=None,
    wintit=None,
    # interactivity
    color_dict=None,
    nlos=None,
    dinc=None,
    connect=None,
):

    # ------------
    # check inputs

    (
        key,
        proj,
        data,
        dataref,
        color_dict,
        nlos,
        connect,
    ) = _plot_diagnostic_check(
        coll=coll,
        key=key,
        # figure
        proj=proj,
        data=data,
        # interactivity
        color_dict=color_dict,
        nlos=nlos,
        connect=connect,
    )

    # ------------
    # prepare data

    dplot = coll.get_diagnostic_dplot(
        key=key,
        optics=optics,
        elements=elements,
    )

    cam = coll.dobj['diagnostic'][key]['optics'][0]
    camref = coll.dobj['camera'][cam]['dgeom']['ref']
    is2d = coll.dobj['camera'][cam]['dgeom']['type'] == '2d'
    if is2d:
        refx, refy = camref
    else:
        refx = camref[0]

    # -------------------------
    # prepare los interactivity

    los = coll.dobj['diagnostic'][key]['los']

    coll2 = coll.__class__()
    for rr in camref:
        coll2.add_ref(key=rr, size=coll.dref[rr]['size'])

    if los is not None:
        los_x, los_y, los_z = coll.sample_rays(
            key=los,
            res=los_res,
            mode='rel',
        )
        los_r = np.hypot(los_x, los_y)
        reflos = coll.dobj['rays'][los]['ref']
        ref_los = (reflos[1:], reflos[1:])

        coll2.add_ref(key=reflos[0], size=los_x.shape[0])
        coll2.add_data(key='los_x', data=los_x, ref=reflos)
        coll2.add_data(key='los_y', data=los_y, ref=reflos)
        coll2.add_data(key='los_z', data=los_z, ref=reflos)
        coll2.add_data(key='los_r', data=los_r, ref=reflos)

    if data is not None:
        if is2d:
            keyx, keyy = coll.dobj['camera'][cam]['dgeom']['cents']

            datax = coll.ddata[keyx]['data']
            datay = coll.ddata[keyy]['data']

            coll2.add_data(key=keyx, data=datax, ref=refx)
            coll2.add_data(key=keyy, data=datay, ref=refy)
        else:
            keyx = 'i0'
            datax = np.arange(0, coll.dref[refx]['size'])
            coll2.add_data(key=keyx, data=datax, ref=refx)

    # -------------------------
    # prepare data interactivity

    reft = None
    if data is not None:
        if dataref is None:
            dataref = coll.ddata[data]['ref']
        if dataref == camref:
            if isinstance(data, str):
                datamap = coll.ddata[data]['data'].T
            else:
                datamap = data.T
        elif len(dataref) == len(camref) + 1:
            dataref == camref
            datamap = coll.ddata[data]['data'][0, ...].T
            # reft = [rr for rr in dataref if rr not in camref][0]

        else:
            raise NotImplementedError()

        if is2d:
            extent = (
                datax[0] - 0.5*(datax[1] - datax[0]),
                datax[-1] + 0.5*(datax[-1] - datax[-2]),
                datay[0] - 0.5*(datay[1] - datay[0]),
                datay[-1] + 0.5*(datay[-1] - datay[-2]),
            )
        else:
            ylab = f"{coll.ddata[data]['quant']} ({coll.ddata[data]['units']})"

    # -----------------
    # prepare figure

    if dax is None:

        dax = _generic_plot.get_dax_diag(
            proj=proj,
            dmargin=dmargin,
            fs=fs,
            wintit=wintit,
            tit=key,
            is2d=is2d,
        )

    dax = _generic_check._check_dax(dax=dax, main=proj[0])

    # -----------------
    # plot static parts

    for k0, v0 in dplot.items():

        for k1, v1 in v0.items():

            # cross
            kax = 'cross'
            if dax.get(kax) is not None:
                ax = dax[kax]['handle']

                if k1.startswith('v-'):
                    ax.quiver(
                        v1['r'],
                        v1['z'],
                        v1['ur'],
                        v1['uz'],
                        **v1.get('props', {}),
                    )

                else:
                    ax.plot(
                        v1['r'],
                        v1['z'],
                        **v1.get('props', {}),
                    )

            # hor
            kax = 'hor'
            if dax.get(kax) is not None:
                ax = dax[kax]['handle']

                if k1.startswith('v-'):
                    ax.quiver(
                        v1['x'],
                        v1['y'],
                        v1['ux'],
                        v1['uy'],
                        **v1.get('props', {}),
                    )

                else:
                    ax.plot(
                        v1['x'],
                        v1['y'],
                        **v1.get('props', {}),
                    )

            # 3d
            kax = '3d'
            if dax.get(kax) is not None:
                ax = dax[kax]['handle']

                if k1.startswith('v-'):
                    ax.quiver(
                        v1['x'],
                        v1['y'],
                        v1['z'],
                        v1['ux'],
                        v1['uy'],
                        v1['uz'],
                        **v1.get('props', {}),
                    )

                else:
                    ax.plot(
                        v1['x'],
                        v1['y'],
                        v1['z'],
                        **v1.get('props', {}),
                    )

            # plotting of 2d camera contour
            kax = 'camera'
            if dax.get(kax) is not None:
                ax = dax[kax]['handle']

                if k0 == cam and is2d and k1 == 'o':
                    ax.plot(
                        v1['x0'],
                        v1['x1'],
                        **v1.get('props', {}),
                    )

    # plot data
    kax = 'camera'
    if dax.get(kax) is not None and data is not None:
        ax = dax[kax]['handle']

        if is2d and reft is None:
            im = ax.imshow(
                datamap,
                extent=extent,
                cmap=cmap,
                vmin=vmin,
                vmax=vmax,
                origin='lower',
                interpolation='nearest',
            )
            plt.colorbar(im, ax=ax)

        elif reft is None:
            ax.plot(
                datamap,
                c='k',
                ls='-',
                lw=1.,
                marker='.',
                ms=6,
            )
            ax.set_xlim(-1, datax[-1] + 1)
            ax.set_ylabel(ylab)

    # ----------------
    # define and set dgroup

    if coll2 is not None:
        dgroup = {
            'x': {
                'ref': [refx],
                'data': ['index'],
                'nmax': nlos,
            },
        }

        if is2d:
            dgroup.update({
                'y': {
                    'ref': [refy],
                    'data': ['index'],
                    'nmax': nlos,
                },
            })

        if reft is not None:
            dgroup.update({
                't': {
                    'ref': [reft],
                    'data': ['index'],
                    'nmax': 1,
                },
            })

    # ------------------
    # plot mobile parts

    if los is not None:

        nan = np.full((los_x.shape[0],), np.nan)

        # cross
        kax = 'cross'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']

            for ii in range(nlos):
                l0, = ax.plot(
                    nan,
                    nan,
                    c=color_dict['x'][ii],
                    ls='-',
                    lw=1.,
                )

                # add mobile
                kl0 = f'los-cross-{ii}'
                coll2.add_mobile(
                    key=kl0,
                    handle=l0,
                    refs=ref_los,
                    data=['los_r', 'los_z'],
                    dtype=['xdata', 'ydata'],
                    axes=kax,
                    ind=ii,
                )

        # hor
        kax = 'hor'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']

            for ii in range(nlos):
                l0, = ax.plot(
                    nan,
                    nan,
                    c=color_dict['x'][ii],
                    ls='-',
                    lw=1.,
                )

                # add mobile
                kl0 = f'los-hor-{ii}'
                coll2.add_mobile(
                    key=kl0,
                    handle=l0,
                    refs=ref_los,
                    data=['los_x', 'los_y'],
                    dtype=['xdata', 'ydata'],
                    axes=kax,
                    ind=ii,
                )

        # 3d
        kax = '3d'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']

            for ii in range(nlos):
                l0, = ax.plot(
                    nan,
                    nan,
                    nan,
                    c=color_dict['x'][ii],
                    ls='-',
                    lw=1.,
                )

                # add mobile
                kl0 = f'los-3d-{ii}'
                # coll2.add_mobile(
                # key=kl0,
                # handle=l0,
                # refs=reflos,
                # data=['index', 'index', 'index'],
                # dtype=['xdata', 'ydata', 'zdata'],
                # axes=kax,
                # ind=ii,
                # )

        # camera
        kax = 'camera'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']

            if is2d:
                for ii in range(nlos):
                    mi, = ax.plot(
                        datax[0:1],
                        datay[0:1],
                        marker='s',
                        ms=6,
                        markeredgecolor=color_dict['x'][ii],
                        markerfacecolor='None',
                    )

                    km = f'm{ii:02.0f}'
                    coll2.add_mobile(
                        key=km,
                        handle=mi,
                        refs=[refx, refy],
                        data=[keyx, keyy],
                        dtype=['xdata', 'ydata'],
                        axes=kax,
                        ind=ii,
                    )

                dax[kax].update(
                    refx=[refx],
                    refy=[refy],
                    datax=keyx,
                    datay=keyy,
                )

            else:
                for ii in range(nlos):
                    lv = ax.axvline(
                        datax[0], c=color_dict['y'][ii], lw=1., ls='-',
                    )
                    kv = f'v{ii:02.0f}'
                    coll2.add_mobile(
                        key=kv,
                        handle=lv,
                        refs=refx,
                        data=keyx,
                        dtype='xdata',
                        axes=kax,
                        ind=ii,
                    )

                dax[kax].update(refx=[refx], datax=keyx)

    # -------
    # config

    if plot_config.__class__.__name__ == 'Config':

        kax = 'cross'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']
            plot_config.plot(lax=ax, proj=kax)

        kax = 'hor'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']
            plot_config.plot(lax=ax, proj=kax)

    # -------
    # connect

    if coll2.dobj.get('mobile') is not None:
        # add axes
        for ii, kax in enumerate(dax.keys()):
            harmonize = ii == len(dax.keys()) - 1
            coll2.add_axes(key=kax, harmonize=harmonize, **dax[kax])

        # connect
        if connect is True:
            coll2.setup_interactivity(
                kinter='inter0',
                dgroup=dgroup,
                dinc=dinc,
            )
            coll2.disconnect_old()
            coll2.connect()

            coll2.show_commands()
            return coll2
        else:
            return coll2, dgroup
    else:
        return dax



# -*- coding: utf-8 -*-


import warnings
import datetime as dtm


import numpy as np
import scipy.interpolate as scpinterp
import matplotlib.pyplot as plt


import Polygon as plg
from Polygon import Utils as plgUtils
import datastock as ds


from . import _class5_reflections_ptsvect
from . import _class5_projections
from . import _class8_compute as _compute


# ##############################################################
# ##############################################################
#           Equivalent aperture
# ##############################################################


def equivalent_apertures(
    # resources
    coll=None,
    key=None,
    pixel=None,
    # inital contour
    add_points=None,
    # options
    convex=None,
    harmonize=None,
    reshape=None,
    return_for_etendue=None,
    # plot
    verb=None,
    plot=None,
    store=None,
):

    # -------------
    # check inputs

    (
        key,
        kref,
        cref,
        ispectro,
        lop_pre,
        lop_pre_cls,
        lop_post,
        lop_post_cls,
        is2d,
        shape0,
        cx,
        cy,
        cz,
        pixel,
        add_points,
        convex,
        harmonize,
        reshape,
        verb,
        plot,
        store,
    ) = _check(
        coll=coll,
        key=key,
        pixel=pixel,
        add_points=add_points,
        convex=convex,
        harmonize=harmonize,
        reshape=reshape,
        verb=verb,
        plot=plot,
        store=store,
    )

    if pixel is None:
        pixel = np.arange(0, cx.size)

    if return_for_etendue is None:
        return_for_etendue = False

    # ---------------
    # Prepare optics

    lpoly_pre = [
        coll.get_optics_poly(
            key=k0,
            add_points=add_points,
            return_outline=False,
        )
        for k0 in lop_pre
    ]
    nop_pre = len(lop_pre)

    lout_post = [
        coll.dobj[c0][k0]['dgeom']['outline']
        for c0, k0 in zip(lop_post_cls, lop_post)
    ]
    lx0_post = [coll.ddata[pp[0]]['data'] for pp in lout_post]
    lx1_post = [coll.ddata[pp[1]]['data'] for pp in lout_post]
    nop_post = len(lop_post)

    # -------------------
    # prepare functions

    coord_x01toxyz = coll.get_optics_x01toxyz(key=kref)
    lcoord_x01toxyz_poly = [
        coll.get_optics_x01toxyz(key=oo)
        for oo in lop_post
    ]

    if len(ispectro) > 0:
        pts2pt = coll.get_optics_reflect_pts2pt(key=kref)
    else:
        pts2pt = None
    ptsvect = coll.get_optics_reflect_ptsvect(key=kref)
    lptsvect_poly = [
        coll.get_optics_reflect_ptsvect(key=oo)
        for oo in lop_post
    ]

    if len(ispectro) == 0:
        func = _get_equivalent_aperture
    else:
        func = _get_equivalent_aperture_spectro

    # -------------------
    # prepare output

    x0 = []
    x1 = []

    # ---------------
    # loop on pixels

    # dt = np.zeros((14,), dtype=float)

    # intial polygon
    p_a = coll.get_optics_outline(key=kref, add_points=False)
    p_a = plg.Polygon(np.array([p_a[0], p_a[1]]).T)

    iok = np.ones((pixel.size,), dtype=bool)

    for ii, ij in enumerate(pixel):

        if verb is True:
            msg = f"\tpixel {ii} / {pixel.size}"
            end = '\n' if ii == len(pixel) - 1 else '\r'
            print(msg, flush=True, end=end)

        p0, p1 = func(
            p_a=p_a,
            pt=np.r_[cx[ij], cy[ij], cz[ij]],
            nop_pre=nop_pre,
            lpoly_pre=lpoly_pre,
            nop_post=nop_post,
            lx0_post=lx0_post,
            lx1_post=lx1_post,
            # functions
            coord_x01toxyz=coord_x01toxyz,
            lcoord_x01toxyz_poly=lcoord_x01toxyz_poly,
            pts2pt=pts2pt,
            ptsvect=ptsvect,
            lptsvect_poly=lptsvect_poly,
            # options
            add_points=add_points,
            convex=convex,
            # timing
            # dt=dt,
        )

        # convex hulli
        if p0 is None or p0.size == 0:
            iok[ii] = False
        elif convex:
            p0, p1 = np.array(plgUtils.convexHull(
                plg.Polygon(np.array([p0, p1]).T)
            ).contour(0)).T

            p0, p1 = _compute._interp_poly(
                lp=[p0, p1],
                add_points=add_points,
                mode='min',
                isclosed=False,
                closed=False,
                ravel=True,
                min_threshold=1.e-5,
            )

        # append
        x0.append(p0)
        x1.append(p1)

    # --------------------
    # harmonize if necessary
    # --------------------

    if harmonize:
        ln = [p0.size if p0 is not None else 0 for p0 in x0]
        nmax = np.max(ln)
        nan = np.full((nmax,), np.nan)
        for ii in range(pixel.size):
            if x0[ii] is None:
                x0[ii] = nan
                x1[ii] = nan
            elif ln[ii] < nmax:
                ndif = nmax - ln[ii]
                irand = np.random.random(ndif)
                irand = irand + np.random.randint(0, ln[ii]-1, ndif)
                imax = np.sort(np.r_[np.arange(0, ln[ii]), irand])
                imax = np.linspace(0, ln[ii]-1, nmax)
                x0[ii] = scpinterp.interp1d(
                    np.arange(0, ln[ii]),
                    x0[ii],
                    kind='linear',
                )(imax)
                x1[ii] = scpinterp.interp1d(
                    np.arange(0, ln[ii]),
                    x1[ii],
                    kind='linear',
                )(imax)

        x0 = np.array(x0)
        x1 = np.array(x1)

    # -------------
    # xyz
    # -------------

    if return_for_etendue is True:

        pts2plane = coll.get_optics_reflect_ptsvect(
            key=kref,
            asplane=True,
        )

        px = np.full(x0.shape, np.nan)
        py = np.full(x0.shape, np.nan)
        pz = np.full(x0.shape, np.nan)
        cents0 = np.full((x0.shape[0],), np.nan)
        cents1 = np.full((x0.shape[0],), np.nan)
        area = np.full((x0.shape[0],), np.nan)

        for ii, ip in enumerate(pixel):

            if not iok[ii]:
                continue

            pxi, pyi, pzi = coord_x01toxyz(x0=x0[ii, :], x1=x1[ii, :])

            (
                px[ii, :], py[ii, :], pz[ii, :],
                _, _, _, _, _, p0, p1,
            ) = pts2plane(
                pts_x=cx[ii],
                pts_y=cy[ii],
                pts_z=cz[ii],
                vect_x=pxi - cx[ii],
                vect_y=pyi - cy[ii],
                vect_z=pzi - cz[ii],
                strict=False,
                return_x01=True,
            )

            # area
            area[ii] = plg.Polygon(np.array([p0, p1]).T).area()

            # centroid in 3d
            cents0[ii], cents1[ii] = plg.Polygon(
                np.array([x0[ii, :], x1[ii, :]]).T
            ).center()

        centsx, centsy, centsz = coord_x01toxyz(
            x0=cents0,
            x1=cents1,
        )

        plane_nin = coll.dobj[cref][kref]['dgeom']['nin']
        spectro = len(ispectro) > 0

    else:
        cents0, cents1 = None, None

    # --------------------
    # reshape if necessary
    # --------------------

    ntot = np.prod(shape0)
    if is2d and harmonize and reshape and x0.shape[0] == ntot:
        shape = tuple(np.r_[shape0, x0.shape[-1]])
        x0 = x0.reshape(shape)
        x1 = x1.reshape(shape)

    # -------------
    # plot & return
    # -------------

    if plot is True:
        out0, out1 = coll.get_optics_outline(key=kref, add_points=False)
        _plot(
            poly_x0=out0,
            poly_x1=out1,
            p0=x0,
            p1=x1,
            cents0=cents0,
            cents1=cents1,
            # options
            tit=f"Reflections on {key}",
            xlab=None,
            ylab=None,
        )

    # timing
    # lt = [f"\t- dt{ii}: {dt[ii]}" for ii in range(len(dt))]
    # print(f"\nTiming for {key}:\n" + "\n".join(lt))

    if return_for_etendue:
        return (
            x0, x1, kref, iok,
            px, py, pz,
            cx, cy, cz,
            centsx, centsy, centsz,
            area, plane_nin, spectro,
        )
    else:
        return x0, x1, kref, iok


# ##############################################################
# ##############################################################
#               check
# ##############################################################


def _check(
    coll=None,
    key=None,
    pixel=None,
    add_points=None,
    convex=None,
    harmonize=None,
    reshape=None,
    verb=None,
    plot=None,
    store=None,
):

    # --------
    # key

    lok = [
        k0 for k0, v0 in coll.dobj.get('diagnostic', {}).items()
        if len(v0['optics']) > 1
    ]
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )

    optics, optics_cls = coll.get_diagnostic_optics(key=key)
    ispectro = [
        ii for ii, cc in enumerate(optics_cls)
        if cc in ['grating', 'crystal']
    ]

    # -------------------------------------------------
    # ldeti: list of individual camera dict (per pixel)

    dgeom = coll.dobj['camera'][optics[0]]['dgeom']
    cx, cy, cz = coll.get_camera_cents_xyz(key=optics[0])
    is2d = dgeom['type'] == '2d'
    shape0 = cx.shape

    if is2d:
        cx = cx.ravel()
        cy = cy.ravel()
        cz = cz.ravel()

    # ---------
    # pixel

    if pixel is not None:
        pixel = np.atleast_1d(pixel).astype(int)

        if pixel.ndim == 2 and pixel.shape[1] == 2 and is2d:
            pixel = pixel[:, 0] * shape0[1]  + pixel[:, 1]

        if pixel.ndim != 1:
            msg = "pixel can only have ndim = 2 for 2d cameras!"
            raise Exception(msg)

    # ------------------------------------
    # compute equivalent optics if spectro

    if len(ispectro) > 0 and len(optics[ispectro[0]+1:]) > 0:

        c0 = (
            len(ispectro) == 1
            or (
                len(ispectro) == 2
                and len(optics[ispectro[1]+1:]) == 0
            )

        )

        # apertures after crystal => reflection
        if c0:
            kref = optics[ispectro[0]]
            cref = optics_cls[ispectro[0]]
            lop_pre = optics[1:ispectro[0]]
            lop_pre_cls = optics_cls[1:ispectro[0]]
            lop_post = optics[ispectro[0]+1:]
            lop_post_cls = optics_cls[ispectro[0]+1:]

        else:
            raise NotImplementedError()

    else:
        kref = optics[-1]
        cref = optics_cls[-1]
        lop_pre = optics[1:-1]
        lop_pre_cls = optics_cls[1:-1]
        lop_post = []
        lop_post_cls = []

    # -----------
    # add_points

    add_points = ds._generic_check._check_var(
        add_points, 'add_points',
        types=int,
        default=3,
        sign='>0',
    )

    # -----------
    # convex

    convex = ds._generic_check._check_var(
        convex, 'convex',
        types=bool,
        default=True,
    )

    # -----------
    # harmonnize

    harmonize = ds._generic_check._check_var(
        harmonize, 'harmonize',
        types=bool,
        default=True,
    )

    # -----------
    # reshape

    reshape = ds._generic_check._check_var(
        reshape, 'reshape',
        types=bool,
        default=True,
    )

    # -----------
    # verb

    verb = ds._generic_check._check_var(
        verb, 'verb',
        types=bool,
        default=True,
    )

    # -----------
    # plot

    plot = ds._generic_check._check_var(
        plot, 'plot',
        default=False,
        types=bool,
    )

    # -----------
    # store

    store = ds._generic_check._check_var(
        store, 'store',
        default=False,
        types=bool,
    )

    return (
        key,
        kref,
        cref,
        ispectro,
        lop_pre,
        lop_pre_cls,
        lop_post,
        lop_post_cls,
        is2d,
        shape0,
        cx,
        cy,
        cz,
        pixel,
        add_points,
        convex,
        harmonize,
        reshape,
        verb,
        plot,
        store,
    )


# ##############################################################
# ##############################################################
#           Equivalent aperture non-spectro
# ##############################################################


def _get_equivalent_aperture(
    p_a=None,
    pt=None,
    nop_pre=None,
    lpoly_pre=None,
    ptsvect=None,
    **kwdargs,
):

    # loop on optics
    for jj in range(nop_pre):

        # project on reference frame
        p0, p1 = ptsvect(
            pts_x=pt[0],
            pts_y=pt[1],
            pts_z=pt[2],
            vect_x=lpoly_pre[jj][0] - pt[0],
            vect_y=lpoly_pre[jj][1] - pt[1],
            vect_z=lpoly_pre[jj][2] - pt[2],
            strict=False,
            return_x01=True,
        )[-2:]

        # inside
        if np.all([p_a.isInside(xx, yy) for xx, yy in zip(p0, p1)]):
            p_a = plg.Polygon(np.array([p0, p1]).T)
        else:
            # intersection
            p_a = p_a & plg.Polygon(np.array([p0, p1]).T)
            if p_a.nPoints() < 3:
                return None, None

    # return
    return np.array(p_a.contour(0)).T


# ##############################################################
# ##############################################################
#           Equivalent aperture spectro
# ##############################################################


def _get_equivalent_aperture_spectro(
    p_a=None,
    pt=None,
    nop_pre=None,
    lpoly_pre=None,
    nop_post=None,
    lx0_post=None,
    lx1_post=None,
    # functions
    coord_x01toxyz=None,
    lcoord_x01toxyz_poly=None,
    pts2pt=None,
    ptsvect=None,
    lptsvect_poly=None,
    # options
    add_points=None,
    convex=None,
    # timing
    dt=None,
):

    # loop on optics before crystal
    for jj in range(nop_pre):

        # project on reference frame
        p0, p1 = ptsvect(
            pts_x=pt[0],
            pts_y=pt[1],
            pts_z=pt[2],
            vect_x=lpoly_pre[jj][0] - pt[0],
            vect_y=lpoly_pre[jj][1] - pt[1],
            vect_z=lpoly_pre[jj][2] - pt[2],
            strict=False,
            return_x01=True,
        )[-2:]

        # inside
        if np.all([p_a.isInside(xx, yy) for xx, yy in zip(p0, p1)]):
            p_a = plg.Polygon(np.array([p0, p1]).T)
        else:
            # intersection
            p_a = p_a & plg.Polygon(np.array([p0, p1]).T)
            if p_a.nPoints() < 3:
                return None, None

    # extract p0, p1
    p0, p1 = np.array(p_a.contour(0)).T

    # loop on optics after crystal
    for jj in range(nop_post):
        # print(f'\t {jj} / {nop_post}')      # DB

        # reflection
        p0, p1 = _class5_projections._get_reflection(
            # inital contour
            x0=p0,
            x1=p1,
            # polygon observed
            poly_x0=lx0_post[jj],
            poly_x1=lx1_post[jj],
            # observation point
            pt=pt,
            add_points=add_points,
            # functions
            coord_x01toxyz=coord_x01toxyz,
            coord_x01toxyz_poly=lcoord_x01toxyz_poly[jj],
            pts2pt=pts2pt,
            ptsvect=ptsvect,
            ptsvect_poly=lptsvect_poly[jj],
            # timing
            # dt=dt,
        )

        if p0 is None:
            # print('\t \t None 0')
            return p0, p1

        if np.all([p_a.isInside(xx, yy) for xx, yy in zip(p0, p1)]):
            pass
        else:
            # convex hull
            if convex:
                p0, p1 = np.array(plgUtils.convexHull(
                    plg.Polygon(np.array([p0, p1]).T)
                ).contour(0)).T

            # plt.figure()
            # plt.plot(
            #     np.array(p_a.contour(0))[:, 0],
            #     np.array(p_a.contour(0))[:, 1], 
            #     '.-k',
            #     p0, p1, '.-r'
            #     )

            # intersection
            p_a = p_a & plg.Polygon(np.array([p0, p1]).T)
            if p_a.nPoints() < 3:
                # print('\t \t None 1')       # DB
                return None, None

            # update
            p0, p1 = np.array(p_a.contour(0)).T
            
            # interpolate
            if jj < nop_post - 1:
                p0, p1 = _compute._interp_poly(
                    lp=[p0, p1],
                    add_points=add_points,
                    mode='min',
                    isclosed=False,
                    closed=False,
                    ravel=True,
                )
                # print(f'\t\t interp => {p0.size} pts')       # DB

    return p0, p1


# ##############################################################
# ##############################################################
#           Plot
# ##############################################################


def _plot(
    poly_x0=None,
    poly_x1=None,
    p0=None,
    p1=None,
    cents0=None,
    cents1=None,
    # graph options
    fs=None,
    tit=None,
    xlab=None,
    ylab=None,
):

    fig = plt.figure(figsize=fs)
    ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])

    ax.set_title(tit)
    ax.set_xlabel(xlab)
    ax.set_ylabel(ylab)

    i0 = np.r_[np.arange(0, poly_x0.size), 0]
    ax.plot(poly_x0[i0], poly_x1[i0], '.-k')
    
    if p0.shape[1] > 0:
        i1 = np.r_[np.arange(0, p0.shape[1]), 0]
        for ii in range(p0.shape[0]):
            ax.plot(p0[ii, i1], p1[ii, i1], '.-', label=f'pix {ii}')

    if cents0 is not None:
        ax.plot(cents0, cents1, 'xr')
        
    ax.legend()
    return



"""
This module contains tests for tofu.geom in its structured version
"""

# Built-in
import os
import shutil
import itertools as itt
import warnings


# Standard
import numpy as np
import matplotlib.pyplot as plt


# tofu-specific
from tofu import __version__
import tofu as tf


_here = os.path.abspath(os.path.dirname(__file__))
VerbHead = 'tofu.mesh.test_01_checks'
_TOFU_USER = os.path.join(os.path.expanduser("~"), '.tofu')
_CUSTOM = os.path.dirname(os.path.dirname(os.path.dirname(_here)))
_CUSTOM = os.path.join(_CUSTOM, 'scripts', 'tofucustom.py')


#######################################################
#
#     Setup and Teardown
#
#######################################################


def setup_module():
    print("Removing user ~/.tofu/ if any")
    if os.path.isdir(_TOFU_USER):
        shutil.rmtree(_TOFU_USER)
    # Recreating clean .tofu
    # out = subprocess.run(_CUSTOM, stdout=PIPE, stderr=PIPE)
    os.system('python '+_CUSTOM)


def teardown_module():
    print("Removing user ~/.tofu/ if any")
    if os.path.isdir(_TOFU_USER):
        shutil.rmtree(_TOFU_USER)


#######################################################
#
#     checking routines
#
#######################################################


class Test01_Inversions():

    @classmethod
    def setup_class(cls):
        pass

    def setup(self):

        # create conf and cam
        conf0 = tf.load_config('WEST-V0')
        cam = tf.geom.utils.create_CamLOS1D(
            pinhole=[3.0, 1., 0.3],
            focal=0.1,
            sensor_size=0.1,
            sensor_nb=30,
            orientation=[-5*np.pi/6, 0, 0],
            config=conf0,
            Name='camH',
            Exp='WEST',
            Diag='SXR',
        )

        # mesh rect deg 1 and 2
        mesh = tf.data.Plasma2D()
        mesh.add_mesh(
            crop_poly=conf0,
            key='m1',
            res=0.10,
            deg=0,
        )
        mesh.add_bsplines(deg=1)
        mesh.add_bsplines(deg=2)

        # add 2d radius for polar mesh
        kR, kZ = mesh.dobj['mesh']['m1']['knots']
        R, Z = mesh.ddata[kR]['data'], mesh.ddata[kZ]['data']
        nR, nZ = R.size, Z.size
        R = np.repeat(R[:, None], nZ, axis=1)
        Z = np.repeat(Z[None, :], nR, axis=0)
        rad2d = ((R-2.4)/0.4)**2 + (Z/0.5)**2
        krad = 'rad2d'
        mesh.add_data(key=krad, data=rad2d, ref='m1-bs1')

        # mesh polar deg 1 and 2
        mesh.add_mesh_polar(
            key='m2',
            radius=[0, 0.1, 0.2, 0.3, 0.4, 0.5, 1.],
            radius2d=krad,
            deg=0,
        )
        mesh.add_bsplines(key='m2', deg=1)
        mesh.add_bsplines(key='m2', deg=2)

        # add geometry matrices
        chan = np.arange(0, 30)
        mesh.add_ref(key='chan', data=chan, group='chan')
        mesh.add_geometry_matrix(cam=cam, key='m1-bs0', key_chan='chan')
        mesh.add_geometry_matrix(cam=cam, key='m1-bs1', key_chan='chan')
        mesh.add_geometry_matrix(cam=cam, key='m1-bs2', key_chan='chan')
        mesh.add_geometry_matrix(cam=cam, key='m2-bs0', key_chan='chan')
        mesh.add_geometry_matrix(cam=cam, key='m2-bs1', key_chan='chan')
        mesh.add_geometry_matrix(cam=cam, key='m2-bs2', key_chan='chan')

        # add data
        t0 = np.array([0])
        t1 = np.array([0, 1.])
        data0 = np.exp(-(chan - 15.)**2/10**2)
        data1 = (
            np.exp(-(chan - 15.)**2/10**2)
            + 0.1*np.cos(t1)[:, None]*np.exp(-(chan - 15)**2/2**2)
        )
        # mesh.add_ref(key='t0', data=t0, units='s', group='time')
        mesh.add_ref(key='t1', data=t1, units='s', dim='time')
        mesh.add_data(key='data0', data=data0, ref=('chan',))
        mesh.add_data(key='data1', data=data1, ref=('t1', 'chan'))

        self.mesh = mesh

    def teardown(self):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def test01_run_all_and_plot(self):

        dalgo = tf.data.get_available_inversions_algo(returnas=dict)
        lstore = [True, False]
        lkdata = ['data0', 'data1']

        # running
        for kmat in self.mesh.dobj['matrix'].keys():

            kbs = self.mesh.dobj['matrix'][kmat]['bsplines']
            km = self.mesh.dobj['bsplines'][kbs]['mesh']
            mtype = self.mesh.dobj['mesh'][km]['type']
            deg = self.mesh.dobj['bsplines'][kbs]['deg']

            if deg in [0, 1]:
                lop = ['D1N2']
            else:
                lop = ['D1N2', 'D2N2']

            for comb in itt.product(dalgo.keys(), lkdata, lop, lstore):

                if comb[2] == 'D2N2' and deg != 2:
                    continue

                if 'mfr' in comb[0].lower() and deg != 0:
                    continue

                algofam = dalgo[comb[0]]['family']
                if algofam == 'Non-regularized' and mtype != 'polar':
                    continue
                if algofam != 'Non-regularized' and mtype == 'polar':
                    continue

                try:
                    self.mesh.add_inversion(
                        algo=comb[0],
                        key_matrix=kmat,
                        key_data=comb[1],
                        sigma=0.10,
                        operator=comb[2],
                        store=comb[3],
                        conv_crit=1.e-3,
                        kwdargs={'tol': 1.e-4},
                        verb=0,
                    )
                    ksig = f'{comb[1]}-sigma'
                    if ksig in self.mesh.ddata.keys():
                        self.mesh.remove_data(ksig)

                except Exception as err:
                    c0 = (
                        dalgo[comb[0]]['source'] == 'tomotok'
                        and comb[2] == 'D1N2'
                        and kmat == 'matrix0'
                    )
                    if c0:
                        # Discrete gradient seem to be not positive-definite
                        # To be investigated...
                        pass
                    else:
                        raise err

        # plotting
        linv = list(self.mesh.dobj['inversions'].keys())[::7]
        for kinv in linv:
            dax = self.mesh.plot_inversion(key=kinv)

        plt.close('all')



# -*- coding: utf-8 -*-


# Built-in
# import copy


# Common
import numpy as np
import datastock as ds


# tofu
from . import _class1_Plasma2D
from . import _class2_check as _check
from . import _class2_compute as _compute
from . import _class2_plot as _plot


__all__ = ['Rays']


# #############################################################################
# #############################################################################
#                           Diagnostic
# #############################################################################


class Rays(_class1_Plasma2D.Plasma2D):

    _show_in_summary = 'all'

    _dshow = dict(_class1_Plasma2D.Plasma2D._dshow)
    _dshow.update({
        'rays': [
            'shape',
            'ref',
            'pts',
            'lamb',
            'Rmin',
            'alpha',
            'dalpha',
            'dbeta',
        ],
    })

    def add_rays(
        self,
        key=None,
        # start points
        start_x=None,
        start_y=None,
        start_z=None,
        # wavelength
        lamb=None,
        # ref
        ref=None,
        # from pts
        pts_x=None,
        pts_y=None,
        pts_z=None,
        # from ray-tracing (vect + length or config or diag)
        vect_x=None,
        vect_y=None,
        vect_z=None,
        length=None,
        config=None,
        reflections_nb=None,
        reflections_type=None,
        diag=None,
    ):
        """ Add a set of rays

        Can be defined from:
            - pts: explicit
            - starting pts + vectors

        """

        # check / format input
        dref, ddata, dobj = _check._rays(
            coll=self,
            key=key,
            # start points
            start_x=start_x,
            start_y=start_y,
            start_z=start_z,
            # wavelength
            lamb=lamb,
            # ref
            ref=ref,
            # from pts
            pts_x=pts_x,
            pts_y=pts_y,
            pts_z=pts_z,
            # from ray-tracing (vect + length or config or diag)
            vect_x=vect_x,
            vect_y=vect_y,
            vect_z=vect_z,
            length=length,
            config=config,
            reflections_nb=reflections_nb,
            reflections_type=reflections_type,
            diag=diag,
        )

        # update dicts
        self.update(dref=dref, ddata=ddata, dobj=dobj)

    # --------------
    #  utilities
    # --------------

    def get_rays_start(
        self,
        key=None,
    ):
        return _check._get_start(coll=self, key=key)

    def get_rays_pts(
        self,
        key=None,
    ):
        return _check._get_pts(coll=self, key=key)

    def get_rays_vect(
        self,
        key=None,
        norm=None,
    ):
        return _check._get_vect(coll=self, key=key, norm=norm)

    # --------------
    # discretizing
    # --------------

    def sample_rays(
        self,
        key=None,
        res=None,
        mode=None,
        concatenate=None,
    ):
        """ Return the sampled rays """

        return _compute._sample(
            coll=self,
            key=key,
            res=res,
            mode=mode,
            concatenate=concatenate,
        )

    # --------------
    # plotting
    # --------------

    def plot_rays(
        self,
        key=None,
        proj=None,
        concatenate=None,
        res=None,
        # config
        plot_config=None,
        # figure
        dax=None,
        dmargin=None,
        fs=None,
        wintit=None,
        # interactivity
        color_dict=None,
        nlos=None,
        dinc=None,
        connect=None,
    ):
        return _plot._plot_rays(
            coll=self,
            key=key,
            proj=proj,
            concatenate=concatenate,
            res=res,
            # config
            plot_config=plot_config,
            # figure
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            wintit=wintit,
            # interactivity
            color_dict=color_dict,
            nlos=nlos,
            dinc=dinc,
            connect=connect,
        )



# -*- coding: utf-8 -*-


import warnings


import numpy as np


import datastock as ds


__all__ = ['compute_los_angles']


# ##################################################################
# ##################################################################
#                       Main
# ##################################################################


def compute_los_angles(
    coll=None,
    key=None,
    kref=None,
    is2d=None,
    # los
    los_x=None,
    los_y=None,
    los_z=None,
    dlos_x=None,
    dlos_y=None,
    dlos_z=None,
    iok=None,
    cx=None,
    cy=None,
    cz=None,
    # for storing los
    config=None,
    length=None,
    reflections_nb=None,
    reflections_type=None,
    **kwdargs,
):

    # ------------
    # check inputs

    # key
    lok = list(coll.dobj.get('diagnostic', {}))
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )

    # klos
    klos = f'{key}-los'

    # ref
    optics = coll.dobj['diagnostic'][key]['optics']
    ref = coll.dobj['camera'][optics[0]]['dgeom']['ref']

    # ------------
    # add los

    cx2, cy2, cz2 = coll.get_camera_cents_xyz(key=optics[0])

    coll.add_rays(
        key=klos,
        start_x=cx2,
        start_y=cy2,
        start_z=cz2,
        vect_x=los_x,
        vect_y=los_y,
        vect_z=los_z,
        ref=ref,
        diag=key,
        config=config,
        length=length,
        reflections_nb=reflections_nb,
        reflections_type=reflections_type,
    )

    coll.set_param(
        which='diagnostic',
        key=key,
        param='los',
        value=klos,
    )

    # ------------
    # for spectro => estimate angle variations
    if dlos_x is not None:

        angmin = np.full(cx.size, np.nan)
        angmax = np.full(cx.size, np.nan)

        ptsvect = coll.get_optics_reflect_ptsvect(key=kref)

        for ii in range(cx.size):
            if not iok[ii]:
                continue
            angles = ptsvect(
                pts_x=cx[ii],
                pts_y=cy[ii],
                pts_z=cz[ii],
                vect_x=dlos_x[ii, ...],
                vect_y=dlos_y[ii, ...],
                vect_z=dlos_z[ii, ...],
                strict=True,
                return_x01=False,
            )[6]

            angmin[ii] = np.nanmin(angles)
            angmax[ii] = np.nanmax(angles)

        if is2d:
            angmin = angmin.reshape(los_x.shape)
            angmax = angmax.reshape(los_x.shape)

        # ddata
        kamin = f'{key}_amin'
        kamax = f'{key}_amax'
        ddata = {
            kamin: {
                'data': angmin,
                'ref': ref,
                'dim': 'angle',
                'quant': 'angle',
                'name': 'alpha',
                'units': 'rad',
            },
            kamax: {
                'data': angmax,
                'ref': ref,
                'dim': 'angle',
                'quant': 'angle',
                'name': 'alpha',
                'units': 'rad',
            },
        }
        coll.update(ddata=ddata)

        coll.set_param(
            which='diagnostic',
            key=key,
            param='amin',
            value=kamin,
        )
        coll.set_param(
            which='diagnostic',
            key=key,
            param='amax',
            value=kamax,
        )



# -*- coding: utf-8 -*-


import warnings
import itertools as itt


import numpy as np
import matplotlib.colors as mcolors
import datastock as ds


# #############################################################################
# #############################################################################
#                           Diagnostics
# #############################################################################


def _diagnostics_check(
    coll=None,
    key=None,
    optics=None,
):

    # ----
    # key

    key = ds._generic_check._obj_key(
        d0=coll.dobj.get('diagnostic', {}), short='diag', key=key,
    )

    # ------
    # optics

    if isinstance(optics, str):
        optics = (optics,)

    lcam = list(coll.dobj.get('camera', {}).keys())
    lap = list(coll.dobj.get('aperture', {}).keys())
    lfilt = list(coll.dobj.get('filter', {}).keys())
    lcryst = list(coll.dobj.get('crystal', {}).keys())
    lgrat = list(coll.dobj.get('grating', {}).keys())
    optics = ds._generic_check._check_var_iter(
        optics, 'optics',
        types_iter=str,
        types=(tuple, list),
        allowed=lcam + lap + lfilt + lcryst + lgrat,
    )
    if isinstance(optics, list):
        optics = tuple(optics)

    # check starts with camera
    if optics[0] not in lcam:
        msg = f"Arg optics must start with a camera!\nProvided: {optics}"
        raise Exception(msg)

    if len(optics) > 1 and any([oo in lcam for oo in optics[1:]]):
        msg = f"Arg optics can only have one camera!\nProvided: {optics}"
        raise Exception(msg)

    # -----------------
    # type of camera

    is2d = coll.dobj['camera'][optics[0]]['dgeom']['type'] == '2d'

    # -------------------------------------------
    # check all optics are on good side of camera

    cam = optics[0]
    dgeom_cam = coll.dobj['camera'][cam]['dgeom']
    last_ref = cam
    last_ref_cls = 'camera'
    for oo in optics[1:]:

        if oo in lap:
            cls = 'aperture'
        elif oo in lfilt:
            cls = 'filter'
        elif oo in lcryst:
            cls = 'crystal'
        else:
            cls = 'grating'

        dgeom = coll.dobj[cls][oo]['dgeom']

        px, py, pz = coll.get_optics_poly(key=oo)

        dgeom_lastref = coll.dobj[last_ref_cls][last_ref]['dgeom']

        if (last_ref == cam and is2d) or last_ref != cam:
            cent = dgeom_lastref['cent']
            nin = dgeom_lastref['nin']

            iout = (
                (px - cent[0])*nin[0]
                + (py - cent[1])*nin[1]
                + (pz - cent[2])*nin[2]
            ) <= 0
            if np.any(iout):
                msg = (
                    f"The following points of aperture '{oo}' are on the wrong"
                    f"side of lastref '{cam}':\n"
                    f"{iout.nonzero()[0]}"
                )
                raise Exception(msg)

        else:
            assert last_ref == cam and not is2d

            cx, cy, cz = dgeom_cam['cents']
            cx = coll.ddata[cx]['data'][None, :]
            cy = coll.ddata[cy]['data'][None, :]
            cz = coll.ddata[cz]['data'][None, :]

            if dgeom_cam['parallel']:
                ninx, niny, ninz = dgeom_cam['nin']
            else:
                ninx, niny, ninz = dgeom_cam['nin']
                ninx = coll.ddata[ninx]['data'][None, :]
                niny = coll.ddata[niny]['data'][None, :]
                ninz = coll.ddata[ninz]['data'][None, :]

            iout = (
                (px[:, None] - cx)*ninx
                + (py[:, None] - cy)*niny
                + (pz[:, None] - cz)*ninz
            ) <= 0
            if np.any(iout):
                msg = (
                    f"The following points of {cls} '{oo}' are on the wrong"
                    f"side of camera '{cam}':\n"
                    f"{np.unique(iout.nonzero()[0])}"
                )
                warnings.warn(msg)

        # update last_ref ?
        if cls in ['crystal', 'grating']:
            last_ref = oo
            last_ref_cls = cls

    # -----------------
    # compute los

    compute = len(optics) > 1

    return key, optics, is2d, compute


def _diagnostics(
    coll=None,
    key=None,
    optics=None,
    **kwdargs,
):

    # ------------
    # check inputs

    key, optics, is2d, compute = _diagnostics_check(
        coll=coll,
        key=key,
        optics=optics,
    )

    # ----------
    # is spectro

    spectro = any([
        k0 in coll.dobj.get('crystal', {}).keys()
        or k0 in coll.dobj.get('grating', {}).keys()
        for k0 in optics
    ])

    # --------
    # dobj

    dobj = {
        'diagnostic': {
            key: {
                'optics': optics,
                'spectro': spectro,
                'etendue': None,
                'etend_type': None,
                'los': None,
                'vos': None,
                'amin': None,
                'amax': None,
            },
        },
    }

    # -----------
    # kwdargs

    if len(kwdargs) > 0:
        for k0, v0 in kwdargs.items():
            if not isinstance(k0, str):
                continue
            elif k0 in dobj['diagnostic'][key].keys():
                continue
            else:
                dobj['diagnostic'][key][k0] = v0

    return None, None, dobj


# ##################################################################
# ##################################################################
#                           get ref
# ##################################################################


def get_ref(coll=None, key=None):

    # ---------
    # check key

    lok = list(coll.dobj.get('diagnostic', {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )

    # -----------
    # return

    cam = coll.dobj['diagnostic'][key]['optics'][0]
    return coll.dobj['camera'][cam]['dgeom']['ref']


# ##################################################################
# ##################################################################
#                       get optics
# ##################################################################


def _get_optics(coll=None, key=None, optics=None):

    # ---------
    # check key

    lcls = ['camera', 'aperture', 'filter', 'crystal', 'grating']
    if optics is None:
        lok = list(coll.dobj.get('diagnostic', {}).keys())
        key = ds._generic_check._check_var(
            key, 'key',
            types=str,
            allowed=lok,
        )
        optics = coll.dobj['diagnostic'][key]['optics']

    else:
        if isinstance(optics, str):
            optics = [optics]
        lok = itt.chain.from_iterable([
            list(coll.dobj.get(cc, {}).keys())
            for cc in lcls
        ])
        optics = ds._generic_check._check_var_iter(
            optics, 'optics',
            types=list,
            types_iter=str,
            allowed=lok,
        )

    # -----------
    # return

    optics_cls = []
    for ii, oo in enumerate(optics):

        lc = [cc for cc in lcls if oo in coll.dobj.get(cc, {}).keys()]
        if len(lc) == 1:
            optics_cls.append(lc[0])

        else:
            msg = f"Diagnostic {key}:"
            if len(lc) == 0:
                msg = f"{msg} no matching class for optics {oo}"
            else:
                msg = f"{msg} multiple matching classes for optics {oo}: {lc}"
            raise Exception(msg)

    return optics, optics_cls


# ##################################################################
# ##################################################################
#                           set color
# ##################################################################


def _set_optics_color(
    coll=None,
    key=None,
    color=None,
):

    # ------------
    # check inputs

    # key
    lk = ['aperture', 'filter', 'crystal', 'grating', 'camera', 'diagnostic']
    dk = {
        k0: list(coll.dobj.get(k0, {}).keys())
        for k0 in lk
    }
    lok = itt.chain.from_iterable([vv for vv in dk.values()])

    if isinstance(key, str):
        key = [key]

    key = ds._generic_check._check_var_iter(
        key, 'key',
        types=(list, tuple),
        types_iter=str,
        allowed=lok,
    )

    # color
    if color is None:
        color = 'k'

    if not mcolors.is_color_like(color):
        msg = (
            f"Arg color for '{key}' must be a matplotlib color!\n"
            f"Provided: {color}\n"
        )
        raise Exception(msg)

    color = mcolors.to_rgba(color)

    # ------------
    # set color

    for k0 in key:
        cls = [k1 for k1, v1 in dk.items() if k0 in v1][0]
        if cls == 'diagnostic':
            for k2 in coll._dobj[cls][k0]['optics']:
                cls2 = [k1 for k1, v1 in dk.items() if k2 in v1][0]
                coll._dobj[cls2][k2]['dmisc']['color'] = color
        else:
            coll._dobj[cls][k0]['dmisc']['color'] = color

    return



# -*- coding: utf-8 -*-


import numpy as np
import scipy.interpolate as scpinterp
import datastock as ds


# ###############################################################
# ###############################################################
#                   sample
# ###############################################################


def _sample(
    coll=None,
    key=None,
    res=None,
    mode=None,
    concatenate=None,
):

    # ------------
    # check inputs

    # res
    res = ds._generic_check._check_var(
        res, 'res',
        types=float,
        default=0.25,
        sign='> 0',
    )

    # mode
    mode = ds._generic_check._check_var(
        mode, 'mode',
        types=str,
        default='rel',
        allowed=['rel', 'abs'],
    )

    # concatenate
    concatenate = ds._generic_check._check_var(
        concatenate, 'concatenate',
        types=bool,
        default=False,
    )
    
    if mode == 'abs':
        concatenate = False

    # --------
    # compute

    pts_x, pts_y, pts_z = coll.get_rays_pts(key=key)

    if mode == 'rel':

        # make sure npts allow to describe all integer indices
        i0 = np.arange(0, pts_x.shape[0])
        npts = i0[-1] * int(np.ceil(1./res))

        N = int(np.ceil((npts - 1) / (i0[-1] - i0[0])))
        npts = N * (i0[-1] - i0[0]) + 1
        i1 = np.linspace(i0[0], i0[-1], npts)

        # interpolate
        pts_x = scpinterp.interp1d(
            i0,
            pts_x,
            kind='linear',
            axis=0,
        )(i1)
        pts_y = scpinterp.interp1d(
            i0,
            pts_y,
            kind='linear',
            axis=0,
        )(i1)
        pts_z = scpinterp.interp1d(
            i0,
            pts_z,
            kind='linear',
            axis=0,
        )(i1)

    else:
        raise NotImplementedError()
        
    # -------------------------------------
    # optional concatenation (for plotting)
    
    if concatenate is True:
        shape = tuple(np.r_[np.r_[1], pts_x.shape[1:]])
        nan = np.full(shape, np.nan)
        pts_x = np.concatenate((pts_x, nan), axis=0).T.ravel()
        pts_y = np.concatenate((pts_y, nan), axis=0).T.ravel()
        pts_z = np.concatenate((pts_z, nan), axis=0).T.ravel()

    return pts_x, pts_y, pts_z



# -*- coding: utf-8 -*-


# Built-in


# Common
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import datastock as ds


# tofu
# from tofu import __version__ as __version__
from . import _generic_check


# #############################################################################
# #############################################################################
#                           inversions
# #############################################################################


def _plot_inversion_check(
    coll=None,
    key=None,
    indbf=None,
    indchan=None,
    cmap=None,
    dcolorbar=None,
    dleg=None,
    dax=None,
    connect=None,
):

    # key
    if 'inversions' not in coll.dobj.keys():
        msg = 'No inversions available!'
        raise Exception(msg)

    lk = list(coll.dobj['inversions'].keys())
    keyinv = ds._generic_check._check_var(
        key, 'key',
        default=None,
        types=str,
        allowed=lk,
    )
    keymat = coll.dobj['inversions'][keyinv]['matrix']
    keydata = coll.dobj['inversions'][keyinv]['data_in']
    keyretro = coll.dobj['inversions'][keyinv]['retrofit']
    keybs = coll.dobj['matrix'][keymat]['bsplines']
    keym = coll.dobj['bsplines'][keybs]['mesh']
    mtype = coll.dobj[coll._which_mesh][keym]['type']
    # refbs = coll.dobj['bsplines'][keybs]['ref']

    crop = coll.dobj['matrix'][keymat]['crop']
    if crop is True:
        cropbs = coll.dobj['bsplines'][keybs]['crop']
        cropbs = coll.ddata[cropbs]['data']
    else:
        cropbs = None

    # cmap
    if cmap is None:
        cmap = 'viridis'

    # dcolorbar
    defdcolorbar = {
        # 'location': 'right',
        'fraction': 0.15,
        'orientation': 'vertical',
    }
    dcolorbar = ds._generic_check._check_var(
        dcolorbar, 'dcolorbar',
        default=defdcolorbar,
        types=dict,
    )

    # dleg
    defdleg = {
        'bbox_to_anchor': (1.1, 1.),
        'loc': 'upper left',
        'frameon': True,
    }
    dleg = ds._generic_check._check_var(
        dleg, 'dleg',
        default=defdleg,
        types=(bool, dict),
    )

    # connect
    connect = ds._generic_check._check_var(
        connect, 'connect',
        default=True,
        types=bool,
    )

    return (
        keyinv, keymat, keybs, keydata, keyretro, mtype,
        cropbs, cmap, dcolorbar, dleg, connect,
    )


def _plot_inversion_prepare(
    coll=None,
    coll2=None,
    mtype=None,
    keyinv=None,
    key_matrix=None,
    key_data=None,
    key_retro=None,
):



    # -----------------
    # add nearest-neighbourg interpolated data

    reft, keyt, time = coll.get_time(key=keyinv)[2:5]
    if coll.get_time(key=key_matrix)[0]:
        keyt_data = coll.get_time(key=key_data)[3]
        if keyt_data != keyt:
            dind = coll.get_time(
                key=key_data,
                t=keyt,
            )[-1]
            data = coll.ddata[key_data]['data'][dind['ind'], :]
        else:
            data = coll.ddata[key_data]['data']
    else:
        data = coll.ddata[key_data]['data']

    # ----------------
    # add data + retro

    # ref
    ref_retro = coll.ddata[key_retro]['ref']

    # chan vector
    chan = coll.get_ref_vector(key=key_data, ref=ref_retro[-1])[4]
    if chan is None:
        chan = np.arange(0, coll.dref[ref_retro[-1]]['size'])

    # add ref
    for rr in ref_retro:
        if rr not in coll2.dref.keys():
            coll2.add_ref(key=rr, size=coll.dref[rr]['size'])

    # data
    coll2.add_data(
        key=key_data,
        data=data,
        ref=ref_retro,
    )

    # retro
    if key_retro not in coll2.ddata.keys():
        coll2.add_data(
            key=key_retro,
            data=coll.ddata[key_retro]['data'],
            ref=ref_retro,
        )

    # err
    key_err = f'{key_data}-err'
    err = coll.ddata[key_retro]['data'] - data
    coll2.add_data(
        key=key_err,
        data=err,
        ref=ref_retro,
    )

    # chi2n, nu, reg, niter


    # # data
    # refchan = coll.ddata[keydata]['ref'][-1]
    # chan = coll.ddata[keychan]['data']
    # hastime = coll.ddata[keydata]['data'].ndim == 2

    # if hastime:
        # data = coll.ddata[keydata]['data']
        # time = np.arange(coll.ddata[keydata]['data'].shape[0])
    # else:
        # data = coll.ddata[keydata]['data'][None, :]
        # time = [0]
        # indt = 0

    # reconstructed data
    # matrix = coll.ddata[keymat]['data']
    # sol = coll.ddata[keyinv]['data']
    # shapebs = coll.dobj['bsplines'][keybs]['shape']
    # nbs = int(np.prod(shapebs))

    # if cropbs is not None:
        # cropbs_flat = cropbs.ravel(order='F')

    # if hastime:
        # nt = sol.shape[0]
        # if sol.ndim == 3:
            # sol_flat = sol.reshape((nt, nbs), order='F')
        # else:
            # sol_flat = sol
        # if cropbs is not None:
            # data_re = matrix.dot(sol_flat[:, cropbs_flat].T)
        # else:
            # data_re = matrix.dot(sol_flat[:, ...].T)
    # else:
        # if sol.ndim == 2:
            # sol_flat = sol.ravel(order='F')
        # else:
            # sol_flat = sol
        # if cropbs is not None:
            # data_re = matrix.dot(sol_flat[cropbs_flat].T)[:, None]
        # else:
            # data_re = matrix.dot(sol_flat.T)[:, None]

    # inversion parameters
    if reft is not None:
        chi2n = coll.ddata[f'{keyinv}-chi2n']['data']
        mu = coll.ddata[f'{keyinv}-mu']['data']
        reg = coll.ddata[f'{keyinv}-reg']['data']
        niter = coll.ddata[f'{keyinv}-niter']['data']
    else:
        chi2n = None    # coll.dobj['inversions'][keyinv]['chi2n']
        mu = None       # coll.dobj['inversions'][keyinv]['mu']
        reg = None      # coll.dobj['inversions'][keyinv]['reg']
        niter = None    # coll.dobj['inversions'][keyinv]['niter']

    datamin = min([np.nanmin(data), np.nanmin(coll2.ddata[key_retro]['data'])])
    datamax = max([np.nanmax(data), np.nanmax(coll2.ddata[key_retro]['data'])])
    errmin = np.nanmin(coll2.ddata[key_err]['data'])
    errmax = np.nanmax(coll2.ddata[key_err]['data'])
    errmax = max(np.abs(errmin), np.abs(errmax))

    return (
        chan, time, reft,
        key_err, chi2n, mu, reg, niter,
        datamin, datamax, errmax,
    )



def plot_inversion(
    coll=None,
    key=None,
    indt=None,
    res=None,
    vmin=None,
    vmax=None,
    cmap=None,
    dax=None,
    dmargin=None,
    fs=None,
    dcolorbar=None,
    dleg=None,
    # interactivity
    dinc=None,
    connect=None,
):

    # ------------
    # check inputs

    (
        keyinv, keymat, keybs, key_data, key_retro, mtype,
        cropbs, cmap, dcolorbar, dleg, connect,
    ) = _plot_inversion_check(
        coll=coll,
        key=key,
        cmap=cmap,
        dcolorbar=dcolorbar,
        dleg=dleg,
        connect=connect,
    )

    # --------------
    # plot - prepare

    if dax is None:

        dax = _plot_inversion_create_axes(
            fs=fs,
            dmargin=dmargin,
            mtype=mtype,
        )

    dax = _generic_check._check_dax(dax=dax, main='matrix')

    # --------------
    # plot profile2d

    coll2, dgroup = coll.plot_profile2d(
        key=key,
        res=res,
        # figure
        vmin=vmin,
        vmax=vmax,
        cmap=cmap,
        dax=dax,
        dmargin=dmargin,
        fs=fs,
        dcolorbar=dcolorbar,
        dleg=dleg,
        # interactivity
        dinc=dinc,
        connect=False,
    )
    dax = coll2.dax

    # ------------
    # prepare data

    (
        chan, time, reft,
        key_err, chi2n, mu, reg, niter,
        datamin, datamax, errmax,
    ) = _plot_inversion_prepare(
        coll=coll,
        coll2=coll2,
        keyinv=keyinv,
        key_matrix=keymat,
        key_data=key_data,
        key_retro=key_retro,
    )

    # ---------
    # plot data

    kax = 'retrofit'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        if reft is None:
            retro = coll2.ddata[key_retro]['data']
            data = coll2.ddata[key_data]['data']
        else:
            retro = coll2.ddata[key_retro]['data'][0, :]
            data = coll2.ddata[key_data]['data'][0, :]

        # plot
        l1, = ax.plot(
            chan,
            retro,
            c=(0.8, 0.8, 0.8),
            ls='-',
            lw=1.,
        )

        l0, = ax.plot(
            chan,
            data,
            c='k',
            ls='None',
            lw=2.,
            marker='.',
        )

        # add mobiles
        if reft is not None:
            kl0 = 'data'
            coll2.add_mobile(
                key=kl0,
                handle=l0,
                refs=(reft,),
                data=[key_data],
                dtype=['ydata'],
                axes=kax,
                ind=0,
            )
            kl1 = 'retrofit'
            coll2.add_mobile(
                key=kl1,
                handle=l1,
                refs=(reft,),
                data=[key_retro],
                dtype=['ydata'],
                axes=kax,
                ind=0,
            )

        ax.set_ylim(min(0, datamin), datamax)
        ax.set_xlim(0, chan.size + 1)

    kax = 'err'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        if reft is None:
            err = coll2.ddata[key_err]['data']
        else:
            err = coll2.ddata[key_err]['data'][0, :]

        l0, = ax.plot(
            chan,
            err,
            c='k',
            ls='-',
            lw=1.,
            marker='.',
        )
        ax.axhline(0, color='k', ls='--')

        # add mobile
        if reft is not None:
            kl0 = 'err'
            coll2.add_mobile(
                key=kl0,
                handle=l0,
                refs=(reft,),
                data=[key_err],
                dtype=['ydata'],
                axes=kax,
                ind=0,
            )

        if np.isfinite(errmax):
            ax.set_ylim(-errmax, errmax)

    # # -------------------------
    # # plot inversion parameters

    if reft is not None:
        kax = 'inv-param'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']
            ax.plot(
                # time,
                chi2n / np.nanmax(chi2n),
                c='k',
                ls='-',
                lw=1.,
                marker='.',
                label='nchi2n',
            )

            ax.plot(
                # time,
                reg / np.nanmax(reg),
                c='b',
                ls='-',
                lw=1.,
                marker='.',
                label='mu*reg',
            )

            # add mobile
            l0 = ax.axvline(time[0], c='k', ls='-', lw=1.)

            # add mobile
            kl0 = 't-par'
            coll2.add_mobile(
                key=kl0,
                handle=l0,
                refs=(reft,),
                data=['index'],
                dtype=['xdata'],
                axes=kax,
                ind=0,
            )

        kax = 'niter'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']
            ax.plot(
                # time,
                niter,
                c='k',
                ls='-',
                lw=1.,
                marker='.',
            )

    # connect
    if connect is True:
        coll2.setup_interactivity(kinter='inter0', dgroup=dgroup, dinc=dinc)
        coll2.disconnect_old()
        coll2.connect()

        coll2.show_commands()
        return coll2
    else:
        return coll2, dgroup


def _plot_inversion_create_axes(
    fs=None,
    dmargin=None,
    mtype=None,
):

    if fs is None:
        fs = (16, 10)

    if dmargin is None:
        dmargin = {
            'left': 0.05, 'right': 0.95,
            'bottom': 0.05, 'top': 0.95,
            'hspace': 0.4, 'wspace': 0.3,
        }

    fig = plt.figure(figsize=fs)
    gs = gridspec.GridSpec(ncols=8, nrows=6, **dmargin)

    # ------------------
    # axes for profile2d

    # axes for image
    ax0 = fig.add_subplot(gs[:4, 2:4], aspect='auto')

    # axes for vertical profile
    ax1 = fig.add_subplot(gs[:4, 4], sharey=ax0)

    # axes for horizontal profile
    ax2 = fig.add_subplot(gs[4:, 2:4], sharex=ax0)

    if mtype == 'polar':
        # axes for traces
        ax7 = fig.add_subplot(gs[:2, :2], sharey=ax2)
        # axes for traces
        ax3 = fig.add_subplot(gs[2:4, :2])
    else:
        ax7 = None
        # axes for traces
        ax3 = fig.add_subplot(gs[:3, :2])


    # axes for text
    ax4 = fig.add_subplot(gs[:3, 5], frameon=False)
    ax5 = fig.add_subplot(gs[3:, 5], frameon=False)
    ax6 = fig.add_subplot(gs[4:, :2], frameon=False)

    # ------------------
    # axes for inversion

    # retrofit
    ax8 = fig.add_subplot(gs[:2, 6:])

    # error
    ax9 = fig.add_subplot(gs[2:4, 6:], sharex=ax8)

    # parameters (chi2, ...)
    ax10 = fig.add_subplot(gs[4, 6:], sharex=ax3)

    # nb of iterations
    ax11 = fig.add_subplot(gs[5, 6:], sharex=ax3)

    # dax
    dax = {
        # data
        'matrix': {'handle': ax0, 'type': 'matrix'},
        'vertical': {'handle': ax1, 'type': 'misc'},
        'horizontal': {'handle': ax2, 'type': 'misc'},
        'traces': {'handle': ax3, 'type': 'misc'},
        # inversion
        'retrofit': {'handle': ax8, 'type': 'misc'},
        'err': {'handle': ax9, 'type': 'misc'},
        'inv-param': {'handle': ax10, 'type': 'misc'},
        'niter': {'handle': ax11, 'type': 'misc'},
        # text
        'textX': {'handle': ax4, 'type': 'text'},
        'textY': {'handle': ax5, 'type': 'text'},
        'textZ': {'handle': ax6, 'type': 'text'},
    }

    if ax7 is not None:
        dax['radial'] = {'handle': ax7, 'type': 'misc'}
    return dax

